import argparse

from smartchangelog.gitcmd import log
from smartchangelog import __version__
from smartchangelog.changelog import Changelog
from smartchangelog.commit import Commit


def main() -> None:
    parser = argparse.ArgumentParser(description="Smart changelog report",
                                     formatter_class=argparse.RawDescriptionHelpFormatter)
    parser.add_argument("-v", "--version", help="print smartchangelog version number", action="version",
                        version=__version__)

    parser.add_argument("-r", "--range", help="revision range (in the same meaning than git log command)")
    parser.add_argument("-g", "--groupby", help="list of criteria", nargs="*")

    args = parser.parse_args()

    gitlog = log(revision_range=args.range)

    changelog = Changelog.parse(gitlog)

    if args.groupby:
        criteria = tuple((Commit.property(criterion) for criterion in args.groupby))
    else:
        criteria = ()

    node = changelog.groupby(*criteria)
    print(node.report())
    exit(0)


if __name__ == "__main__":
    main()



from smartchangelog import datetools
from smartchangelog.commit import Commit
from smartchangelog.commitmsg import CommitType
from tests.unit import data_file_path


class TestCommit:
    def test_parse(self):
        # GIVEN
        with open(data_file_path('one.gitlog'), encoding='utf-8') as log_file:
            log = log_file.read()
        expected = Commit(
            id='a6f79b56acbb9e58327ecf91feed611bb614927f',
            author='Nicolas Gouzy <nicolas.gouzy@orange.com>',
            date=datetools.str2date('2017-03-23 17:30:56 +0100'),
            type=CommitType.refactor,
            scope='changelog',
            subject='better model',
            body='NamedTuple rocks !'
        )
        # WHEN
        changelog_item = Commit.parse(log)
        # THEN
        assert changelog_item == expected

    def test_strip_lines(self):
        # GIVEN
        string = """
            Lorem ipsum dolor sit amet, consectetur adipiscing elit.
            Phasellus non erat imperdiet, pellentesque nibh et, porta velit.

                Fusce sit amet elit ac magna congue accumsan sed ut tellus.
            Nullam at velit tincidunt, sodales mi quis, gravida metus.


            Quisque pellentesque ipsum nec nunc vehicula tincidunt.
        """
        expected = "Lorem ipsum dolor sit amet, consectetur adipiscing elit.\n" \
                   "Phasellus non erat imperdiet, pellentesque nibh et, porta velit.\n" \
                   "\n" \
                   "Fusce sit amet elit ac magna congue accumsan sed ut tellus.\n" \
                   "Nullam at velit tincidunt, sodales mi quis, gravida metus.\n" \
                   "\n" \
                   "\n" \
                   "Quisque pellentesque ipsum nec nunc vehicula tincidunt."
        # WHEN
        actual = Commit.strip_lines(string)
        # THEN
        assert actual == expected

    def test_property_name(self):
        # GIVEN
        prop = Commit.author
        # WHEN
        property_name = Commit.property_name(prop)
        # THEN
        assert property_name == 'author'


import inspect

from smartchangelog import commitmsg
from mypy import api

from smartchangelog import changelog


def test_static_type_check_with_mypy():
    commitmsg_file = inspect.getfile(commitmsg)
    changelog_file = inspect.getfile(changelog)
    params = ['--ignore-missing-imports', commitmsg_file, changelog_file]
    result = api.run(params)
    if result[0]:
        # FIXME: begin: There are bugs in mypy
        #   * support iteration on enums see https://github.com/python/mypy/issues/2305
        #   * support NamedTuple
        # So, we have to remove irrelevant errors
        check_type_errors = "\n".join(
            (error for error in result[0].strip().split("\n") if error.split("error: ")[1] not in (
                '"CommitType" expects no type arguments, but 1 given',
                'Invalid type "commit_type_str"',
                'Iterable expected',
                '"CommitType" has no attribute "__iter__"',
                'Right hand side values are not supported in NamedTuple',
                'Invalid statement in NamedTuple definition; expected "field_name: field_type"'
            )))
        # FIXME: end
        if len(check_type_errors) > 0:
            raise (Exception(check_type_errors))
    if result[1]:
        raise (Exception(result[1]))



#!/usr/bin/env python3
"""
Git commit hook:
 .git/hooks/commit-msg
"""

import argparse

from smartchangelog.commitmsg import CommitMsg, CommitSyntaxError
from smartchangelog import __version__
from smartchangelog.githook import install, uninstall


def main() -> None:
    parser = argparse.ArgumentParser(description="Git commit message checker",
                                     formatter_class=argparse.RawDescriptionHelpFormatter,
                                     epilog=CommitMsg.help())
    parser.add_argument("-v", "--version", help="print commit-msg version number", action="version",
                        version=__version__)

    group = parser.add_mutually_exclusive_group()
    group.add_argument("msg", help="the commit message to check", nargs="?")
    group.add_argument("-i", "--install_hook", action="store_true")
    group.add_argument("-u", "--uninstall_hook", action="store_true")

    args = parser.parse_args()

    if args.install_hook:
        hook_path = install()
        if hook_path:
            print("commit-msg hook installed in {path}".format(path=hook_path))
    elif args.uninstall_hook:
        hook_path = uninstall()
        if hook_path:
            print("commit-msg hook removed from {path}".format(path=hook_path))
    else:
        msg = args.msg
        if "COMMIT_EDITMSG" in msg:
            with open(args.msg) as msg_file:
                msg = msg_file.read()
        try:
            CommitMsg.parse(msg)
        except CommitSyntaxError as e:
            parser.error("{error}\n\n{help}".format(error=e, help=CommitMsg.help()))

    exit(0)


if __name__ == "__main__":
    main()



import inspect
import os
import sys
from contextlib import contextmanager
from io import StringIO

from typing import Iterator, TextIO, cast

from smartchangelog.scripts import commitmsg_script

"""Path of the file containing commitmsg_script.py file"""
commitmsg_script_path = inspect.getfile(commitmsg_script)


@contextmanager
def set_commit_editmsg(msg: str) -> Iterator[TextIO]:
    filename = 'COMMIT_EDITMSG'
    with open(filename, mode='w') as f:
        f.write(msg)
    try:
        yield cast(TextIO, f)
    finally:
        if os.path.isfile(filename):
            os.remove(filename)


@contextmanager
def set_args(*args):
    old = list(sys.argv)
    sys.argv[:] = args
    oldout, olderr = sys.stdout, sys.stderr
    sys.stdout, sys.stderr = StringIO(), StringIO()
    try:
        yield sys.stdout, sys.stderr
    finally:
        sys.stdout.seek(0)
        sys.stderr.seek(0)
        sys.argv[:] = old
        sys.stdout, sys.stderr = oldout, olderr



from smartchangelog import datetools
from smartchangelog.changelog import Changelog, Node
from smartchangelog.commit import Commit
from smartchangelog.commitmsg import CommitType
from tests.unit import data_file_path


class TestChangelog:
    def test_parse(self):
        # GIVEN
        with open(data_file_path('big.gitlog'), encoding='utf-8') as log_file:
            log = log_file.read()
        expected_commit_with_scope = Commit(
            id='a6f79b56acbb9e58327ecf91feed611bb614927f',
            author='Nicolas Gouzy <nicolas.gouzy@orange.com>',
            date=datetools.str2date('2017-03-23 17:30:56 +0100'),
            type=CommitType.refactor,
            scope='changelog',
            subject="better model",
            body='NamedTuple rocks !',
            footer=None
        )
        expected_commit_without_scope = Commit(
            id='597ec5676235e18f5a607726603df944da5be7fe',
            author='Nicolas Gouzy <nicolas.gouzy@orange.com>',
            date=datetools.str2date('2017-03-22 15:28:45 +0100'),
            type=None,
            scope=None,
            subject='Merge branch develop into master',
            body=None,
            footer=None
        )
        # WHEN
        changelog = Changelog.parse(log)
        # THEN
        assert (len(changelog) == 35)
        assert changelog[0] == expected_commit_with_scope
        assert changelog[1] == expected_commit_without_scope

    def test_groupby(self):
        # GIVEN
        with open(data_file_path('big.gitlog'), encoding='utf-8') as log_file:
            log = log_file.read()
        changelog = Changelog.parse(log)
        # WHEN
        node = changelog.groupby(Commit.type, Commit.scope)
        # THEN
        assert len(node) == len(changelog)


class TestNode:
    def test_len_with_empty_tree(self):
        # GIVEN
        tree = Node()
        # WHEN
        # THEN
        assert len(tree) == 1

    def test_len_with_small_tree(self):
        # GIVEN
        children = tuple([Node(name=str(i)) for i in range(10)])
        tree = Node(children=children)
        # WHEN
        actual = len(tree)
        # THEN
        assert actual == 10

    def test_len_with_tree(self):
        # GIVEN
        children = tuple([Node(name=str(i), children=tuple([Node(), Node()])) for i in range(10)])
        tree = Node(children=children)
        # WHEN
        actual = len(tree)
        # THEN
        assert actual == 20

    def test_report_with_big_git_log(self):
        # GIVEN
        with open(data_file_path('big.gitlog'), encoding='utf-8') as log_file:
            log = log_file.read()
        changelog = Changelog.parse(log)
        node = changelog.groupby(Commit.type, Commit.scope)
        with open(data_file_path('big.md'), encoding='utf-8') as md_file:
            expected = md_file.read()
        # WHEN
        report = node.report()
        # THEN
        assert report == expected



import os
import pytest

from smartchangelog.gitcmd import GitCmdError, is_inside_work_tree, get_gitdir, tag
from tests.unit import data_dir_path


@pytest.fixture(scope='function')
def cmd():
    old_cmd = os.getcwd()
    yield None
    os.chdir(old_cmd)


@pytest.mark.usefixtures('cmd')
def test_is_inside_work_tree_ok():
    # GIVEN
    os.chdir(data_dir_path())
    # WHEN
    result = is_inside_work_tree()
    # THEN
    assert result


@pytest.mark.usefixtures('cmd')
def test_is_inside_work_tree_ko():
    # GIVEN
    os.chdir(os.path.expanduser('~'))
    # WHEN
    result = is_inside_work_tree()
    # THEN
    assert not result


@pytest.mark.usefixtures('cmd')
def test_get_gitdir_ok():
    # GIVEN
    os.chdir(data_dir_path())
    # WHEN
    gitdir_path = get_gitdir()
    # THEN
    assert os.path.split(gitdir_path)[-1] == '.git'


@pytest.mark.usefixtures('cmd')
def test_get_gitdir_ko():
    # GIVEN
    os.chdir(os.path.expanduser('~'))
    # WHEN
    with pytest.raises(GitCmdError):
        get_gitdir()
        # THEN
        pass


@pytest.mark.usefixtures('cmd')
def test_get_tag():
    # GIVEN
    os.chdir(data_dir_path())
    # WHEN
    tags = tag()
    # THEN
    assert len(tags) > 0




from datetime import datetime, timezone, timedelta

from smartchangelog import datetools


def test_str2date():
    # GIVEN
    expected = datetime(
        year=2017,
        month=3,
        day=21,
        hour=16,
        minute=9,
        second=13,
        tzinfo=timezone(timedelta(hours=1))
    )
    string = '2017-03-21 16:09:13 +0100'
    # WHEN
    date = datetools.str2date(string)
    # THEN
    assert date == expected


def test_date2str():
    # GIVEN
    expected = '2017-03-21 16:09:13 +0100'
    dt = datetime(
        year=2017,
        month=3,
        day=21,
        hour=16,
        minute=9,
        second=13,
        tzinfo=timezone(timedelta(hours=1))
    )
    # WHEN
    string = datetools.date2str(dt)
    # THEN
    assert string == expected



import re
from collections import Iterable
from io import StringIO
from itertools import groupby

from typing import List, Tuple, Callable, Any, IO, cast

from smartchangelog import datetools
from smartchangelog.commit import Commit


class Node:
    def __init__(self, name: str = None, criterion: property = None, children: Tuple['Node'] = None,
                 value: Commit = None) -> None:
        self._parent: 'Node' = None
        self.name = name
        self.criterion = criterion
        self._children: Tuple['Node'] = None
        self.children = children
        self.value = value

    @property
    def parent(self) -> 'Node':
        return self._parent

    @property
    def children(self) -> Tuple['Node']:
        return self._children

    @children.setter
    def children(self, children: Tuple['Node']) -> None:
        if children is not None:
            for node in children:
                node._parent = self
        self._children = children

    def depth_level(self) -> int:
        if self.parent is None:
            return 0
        else:
            return self.parent.depth_level() + 1

    def __len__(self):
        if not self.children:
            return 1
        nb_children = 0
        for child in self.children:
            nb_children += len(child)
        return nb_children

    @classmethod
    def print_multilines(cls, name: str, value: str, file: IO):
        if value:
            lines = value.split('\n')
            if len(lines) == 1:
                print("    * {name}: {value}".format(name=name, value=value), file=file)
            else:
                print("    * {name}:".format(name=name), file=file)
                for line in lines:
                    print("        - {line}".format(line=line), file=file)

    @classmethod
    def print_leaf(cls, commit: Commit, file: IO) -> None:
        print("* subject: {subject}".format(subject=commit.subject or ''), file=file)
        cls.print_multilines(name='body', value=commit.body, file=file)
        cls.print_multilines(name='footer', value=commit.footer, file=file)
        print("    * date: {date}".format(date=datetools.date2str(commit.date)), file=file)
        print("    * author: {author}".format(author=commit.author), file=file)
        print("    * commit: {id}".format(id=commit.id), file=file)

    def print_header(self, node: 'Node', file: IO):
        print(
            "{header} {criterion_name}: {name}".format(
                header="#" * (self.depth_level() + 1),
                criterion_name=Commit.property_name(node.criterion),
                name=node.name
            ),
            file=file
        )
        print(file=file)

    def report(self) -> str:
        sio = StringIO()
        with sio:
            if self.children is None:
                self.print_leaf(commit=self.value, file=sio)
            else:
                for node in self.children:
                    if node.name:
                        self.print_header(node=node, file=sio)
                    print(node.report().strip('\n'), file=sio)
                    print(file=sio)
            string = sio.getvalue()
            return string


class Changelog(List[Commit]):
    @classmethod
    def parse(cls, log: str) -> 'Changelog':
        raw_commits = re.findall('(commit [a-z0-9]{40}\n(?:.|\n)*?)(?=commit [a-z0-9]{40}|$)', log)
        return Changelog([Commit.parse(rc) for rc in raw_commits])

    def groupby(self, *criteria: property) -> Node:
        if len(criteria) == 0:
            # Sort
            date_prop = cast(property, Commit.date)
            date_getter = cast(Callable[[Commit], Any], date_prop.fget)
            self.sort(key=date_getter)
            return self.node()

        criteria_list = list(criteria)
        criterion = criteria_list.pop(0)
        criterion_getter = cast(Callable[[Commit], Any], criterion.fget)

        # Filter
        # noinspection PyTypeChecker
        categorized_changelog = Changelog([commit for commit in self if criterion_getter(commit) is not None])
        # noinspection PyTypeChecker
        uncategorized_commits = Changelog([commit for commit in self if criterion_getter(commit) is None])

        # Sort
        categorized_changelog.sort(key=criterion_getter)

        # Arrange
        raw_result = self.groupby_to_list(groupby(iterable=categorized_changelog, key=criterion_getter))
        children_list: List[Node] = []
        for key, group in raw_result:
            cl = Changelog(group)
            children_list.append(Node(name=str(key), criterion=criterion, children=cl.groupby(*criteria_list).children))
        if len(uncategorized_commits) > 0:
            children_list.append(uncategorized_commits.node(name="unknown", criterion=criterion))
        children = cast(Tuple[Node], tuple(children_list))

        return Node(children=children)

    def node(self, name: str=None, criterion: property=None) -> Node:
        # noinspection PyTypeChecker
        children = cast(Tuple[Node], tuple(Node(value=commit) for commit in self))
        return Node(name=name, criterion=criterion, children=children)

    @classmethod
    def groupby_to_list(cls, iterable: Iterable):
        return [[key, [i for i in group]] for key, group in iterable]












import os
import tempfile

import pytest

from smartchangelog import githook
from smartchangelog.gitcmd import git_command, GitCmdError


@pytest.fixture(scope='function')
def temp_dir():
    temporary_directory = tempfile.mkdtemp()
    old_cwd = os.getcwd()
    os.chdir(temporary_directory)

    git_command('init')
    git_command('config', 'user.name', 'Nicolas Gouzy')
    git_command('config', 'user.email', 'nicolas.gouzy@gmail.com')

    githook.install()

    sample_file_path = os.path.join(temporary_directory, "sample_file.txt")
    with open(sample_file_path, mode="w") as sample_file:
        sample_file.write("sample content")

    git_command('add', '.')

    yield None

    os.chdir(old_cwd)


@pytest.mark.usefixtures("temp_dir")
def test_git_commit_with_right_msg():
    # GIVEN
    # WHEN
    result = git_command('commit', '-m', 'feat(ui): sample')
    # THEN
    assert result


@pytest.mark.usefixtures("temp_dir")
def test_git_commit_with_wrong_msg():
    # GIVEN
    # WHEN
    with pytest.raises(GitCmdError):
        git_command('commit', '-m', 'wrong commit message')
        # THEN
        pass





import subprocess
import os

from typing import cast, List


class GitCmdError(Exception):
    """
    Git command error
    """


def git_command(*git_args: str) -> str:
    args = ['git'] + cast(List[str], list(git_args))
    cp = subprocess.run(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    if cp.returncode == 0 and len(cp.stderr) == 0:
        return cp.stdout.decode('utf-8').strip('\n')
    else:
        raise GitCmdError(cp.stderr.decode('utf-8').strip('\n'))


def is_inside_work_tree() -> bool:
    try:
        result = git_command('rev-parse', '--is-inside-work-tree')
        return result == 'true'
    except GitCmdError:
        return False


def get_gitdir() -> str:
    if is_inside_work_tree():
        path = os.path.join(git_command('rev-parse', '--show-toplevel'), '.git')
        return os.path.abspath(path)
    else:
        raise GitCmdError("You have to be inside a git work tree")


def log(revision_range: str) -> str:
    return git_command("log", revision_range, "--date", "iso")


def tag() -> List[str]:
    return git_command("tag").split("\n")



import re
from datetime import datetime

from typing import NamedTuple, cast

from smartchangelog import datetools
from smartchangelog.commitmsg import CommitType, CommitMsg, CommitSyntaxError


class _Commit(NamedTuple):
    id: str
    author: str
    date: datetime
    type: CommitType = None
    scope: str = None
    subject: str = None
    body: str = None
    footer: str = None


class Commit(_Commit):
    class Message(NamedTuple):
        type: CommitType = None
        scope: str = None
        subject: str = None
        body: str = None
        footer: str = None

    @classmethod
    def parse(cls, commit: str) -> 'Commit':
        m = re.match('commit (?P<id>[a-z0-9]{40})(?:\n|.)+Author: (?P<author>.*)(?:\n|.)+'
                     'Date: (?P<date>.*)(?P<message>(.|\n)*)',
                     commit)
        gd = m.groupdict()
        message = cls.parse_message(gd['message'])
        commit_id = gd['id']
        author = gd['author']
        date = datetools.str2date(gd['date'].strip())
        return cls(
            id=commit_id,
            author=author,
            date=date,
            type=message.type,
            scope=message.scope,
            subject=message.subject,
            body=message.body,
            footer=message.footer
        )

    @classmethod
    def strip_lines(cls, string) -> str:
        return "\n".join(line.strip() for line in string.strip(' \n').split('\n'))

    @classmethod
    def parse_message(cls, message: str) -> Message:
        message = cls.strip_lines(message)
        try:
            cm = CommitMsg.parse(message)
            return cls.Message(**cm.__dict__)
        except CommitSyntaxError:
            message = re.sub("\n+", "\n", message)
            lines = message.split('\n', maxsplit=1)
            subject = lines[0] or None
            body = None
            if len(lines) > 1:
                body = lines[1] or None
            return cls.Message(
                type=None,
                scope=None,
                subject=subject,
                body=body,
                footer=None
            )

    @classmethod
    def property_name(cls, prop: property) -> str:
        # fixme: change implementation, use _Commit.__dict__
        i = int(prop.__doc__.split(' ')[-1])
        return tuple(cls._fields)[i]

    @classmethod
    def property(cls, name: str):
        prop = cast(property, _Commit.__dict__[name])
        return prop



import os
import shutil


git_path = os.path.join('.git')
hooks_path = os.path.join(git_path, 'hooks')
commitmsg_hook_path = os.path.join(hooks_path, 'commit-msg')


def install():
    check_git_path()
    if not os.path.isdir(hooks_path):
        os.makedirs(hooks_path, mode=0o755, exist_ok=True)
    uninstall()
    commitmsg_script_path = shutil.which('commit-msg')
    assert commitmsg_script_path
    os.symlink(commitmsg_script_path, commitmsg_hook_path)
    assert os.path.exists(commitmsg_hook_path)
    return commitmsg_hook_path


def uninstall():
    check_git_path()
    if os.path.exists(commitmsg_hook_path):
        os.remove(commitmsg_hook_path)
        return commitmsg_hook_path


def check_git_path():
    assert os.path.isdir(git_path)



from datetime import datetime

date_format = "%Y-%m-%d %H:%M:%S %z"


def str2date(string: str) -> datetime:
    return datetime.strptime(string, date_format)


def date2str(dt: datetime) -> str:
    return dt.strftime(date_format)


import pytest

from smartchangelog.commitmsg import CommitMsg, CommitSyntaxError, CommitType


class TestCommitMsg:
    class TestParseFirstLine:
        def test_with_type_and_scope_and_subject(self):
            # GIVEN
            firstline = "feat(ui main): add button"
            # WHEN
            parsed_firstline = CommitMsg.parse_firstline(firstline)
            # THEN
            assert parsed_firstline.type == CommitType.feat
            assert parsed_firstline.scope == "ui main"
            assert parsed_firstline.subject == "add button"

        def test_with_firstline_with_type_and_subject_but_without_scope(self):
            # GIVEN
            msg = "fix: commit-msg hook exit"
            # WHEN
            firstline = CommitMsg.parse_firstline(msg)
            # THEN
            assert firstline.type == CommitType.fix
            assert firstline.scope is None
            assert firstline.subject == "commit-msg hook exit"

        def test_with_wrong_firstline_format(self):
            # GIVEN
            firstline = "bad message"
            with pytest.raises(CommitSyntaxError):
                # WHEN
                CommitMsg.parse_firstline(firstline)
                # THEN CommitSyntaxError is raised

        def test_with_unknown_type(self):
            # GIVEN
            firstline = "unknown(ui): add button"
            with pytest.raises(CommitSyntaxError):
                # WHEN
                CommitMsg.parse_firstline(firstline)
                # THEN CommitSyntaxError is raised

        def test_with_too_long_firstline_length(self):
            # GIVEN
            firstline = "feat(ui): " + "a" * (CommitMsg.FIRSTLINE_MAX_LENGTH + 1)
            with pytest.raises(CommitSyntaxError):
                # WHEN
                CommitMsg.parse_firstline(firstline)
                # THEN CommitSyntaxError is raised

    class TestParseBody:
        def test_with_too_long_body_line_length(self):
            # GIVEN
            body = "body\n" + \
                   "b" * (CommitMsg.BODY_MAX_LENGTH + 1)
            with pytest.raises(CommitSyntaxError):
                # WHEN
                CommitMsg.parse_body(body)
                # THEN CommitSyntaxError is raised

        def test_with_one_line_body(self):
            # GIVEN
            body = "body"
            # WHEN
            actual = CommitMsg.parse_body(body)
            # THEN
            assert actual == body

        def test_with_multi_line_body(self):
            # GIVEN
            body = "first line body\n" + \
                   "second line body"
            # WHEN
            actual = CommitMsg.parse_body(body)
            # THEN
            assert actual == body

    class TestParseFooter:
        def test_with_too_long_footer_line_length(self):
            # GIVEN
            footer = "footer\n" + \
                     "f" * (CommitMsg.FOOTER_MAX_LENGTH + 1)
            with pytest.raises(CommitSyntaxError):
                # WHEN
                CommitMsg.parse_footer(footer)
                # THEN CommitSyntaxError is raised

        def test_with_one_line_footer(self):
            # GIVEN
            footer = "footer"
            # WHEN
            actual = CommitMsg.parse_footer(footer)
            # THEN
            assert actual == footer

        def test_with_multi_line_footer(self):
            # GIVEN
            footer = "first line footer\n" + \
                     "second line footer"
            # WHEN
            actual = CommitMsg.parse_footer(footer)
            # THEN
            assert actual == footer

    class TestParse:
        def test_with_firstline_but_without_body_and_footer(self):
            # GIVEN
            msg = "feat: add button"
            # WHEN
            commit_msg = CommitMsg.parse(msg)
            # THEN
            assert commit_msg.type == CommitType.feat
            assert commit_msg.scope is None
            assert commit_msg.subject == "add button"
            assert commit_msg.body is None
            assert commit_msg.footer is None

        def test_with_firstline_and_body_but_without_footer(self):
            # GIVEN
            msg = "" + \
                  "feat(ui): add button\n" + \
                  "\n" + \
                  "body first line\n" + \
                  "body second line"
            # WHEN
            commit_msg = CommitMsg.parse(msg)
            # THEN
            assert commit_msg.type == CommitType.feat
            assert commit_msg.scope == "ui"
            assert commit_msg.subject == "add button"
            assert commit_msg.body == "body first line\nbody second line"
            assert commit_msg.footer is None

        def test_with_firstline_and_body_and_footer(self):
            # GIVEN
            msg = "feat(ui): add button\n" + \
                  "\n" + \
                  "body\n" + \
                  "\n" + \
                  "footer"
            # WHEN
            commit_msg = CommitMsg.parse(msg)
            # THEN
            assert commit_msg.type == CommitType.feat
            assert commit_msg.scope == "ui"
            assert commit_msg.subject == "add button"
            assert commit_msg.body == "body"
            assert commit_msg.footer == "footer"

        def test_wrong_msg_with_bad_separator_between_firstline_and_body(self):
            # GIVEN
            msg = "feat(ui): add button\n" + \
                  "body"
            with pytest.raises(CommitSyntaxError) as e:
                # WHEN
                CommitMsg.parse(msg)
                # THEN CommitSyntaxError is raised

    class TestEquality:
        def test_equality_with_same_commitmsg(self):
            # GIVEN
            cm1 = CommitMsg(
                msg_type=CommitType.feat,
                scope='conso',
                subject='OEM-372',
                body='add field for nbAlerts'
            )
            cm2 = CommitMsg(
                msg_type=CommitType.feat,
                scope='conso',
                subject='OEM-372',
                body='add field for nbAlerts'
            )
            # WHEN
            # THEN
            assert cm1 == cm2

        def test_equality_with_other_commitmsg(self):
            # GIVEN
            cm1 = CommitMsg(
                msg_type=CommitType.feat,
                scope='conso',
                subject='OEM-372',
                body='add field for nbAlerts'
            )
            cm2 = CommitMsg(
                msg_type=CommitType.fix,
                scope='conso',
                subject='OEM-372',
                body='add field for nbAlerts'
            )
            # WHEN
            # THEN
            assert cm1 != cm2

        def test_equality_with_other_class(self):
            # GIVEN
            cm = CommitMsg(
                msg_type=CommitType.feat,
                scope='conso',
                subject='OEM-372',
                body='add field for nbAlerts'
            )
            s = "a string"
            # WHEN
            # THEN
            assert cm != s


class TestCommitType:
    def test_str(self):
        # GIVEN
        ct = CommitType.feat
        # WHEN
        string = str(ct)
        # THEN
        assert string == 'feat'

    def test_lt_with_commit_type(self):
        # GIVEN
        ct1 = CommitType.feat
        ct2 = CommitType.refactor
        # WHEN
        # THEN
        assert ct1 < ct2

    def test_lt_with_other_class(self):
        # GIVEN
        ct = CommitType.feat
        s = "a string"
        # WHEN
        # THEN
        with pytest.raises(TypeError):
            assert ct < s




import inspect
import os


def data_dir_path() -> str:
    return os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))


def data_file_path(filename: str) -> str:
    return os.path.join(data_dir_path(), 'data', filename)



import pytest

import smartchangelog.scripts.commitmsg_script
from smartchangelog.tools import set_args, set_commit_editmsg, commitmsg_script_path


def test_help_arg():
    # GIVEN
    with set_args(commitmsg_script_path, "-h"), pytest.raises(SystemExit) as e:
        # WHEN
        smartchangelog.scripts.commitmsg_script.main()
    # THEN
    assert e.value.code == 0


def test_right_msg_arg():
    # GIVEN
    with set_args(commitmsg_script_path, 'feat(ui): add button'), pytest.raises(SystemExit) as e:
        # WHEN
        smartchangelog.scripts.commitmsg_script.main()
    # THEN
    assert e.value.code == 0


def test_wrong_msg_arg():
    # GIVEN
    with set_args(commitmsg_script_path, 'wrong commit message'), pytest.raises(SystemExit) as e:
        # WHEN
        smartchangelog.scripts.commitmsg_script.main()
    # THEN
    assert e.value.code != 0


def test_right_msg_file():
    # GIVEN
    with set_commit_editmsg('feat(ui): add button') as f, \
         set_args(commitmsg_script_path, f.name), \
         pytest.raises(SystemExit) as e:
        # WHEN
        smartchangelog.scripts.commitmsg_script.main()
    # THEN
    assert e.value.code == 0


def test_wrong_msg_file():
        # GIVEN
        with set_commit_editmsg('bad format') as f, \
             set_args(commitmsg_script_path, f.name), \
             pytest.raises(SystemExit) as e:
            # WHEN
            smartchangelog.scripts.commitmsg_script.main()
        # THEN
        assert e.value.code != 0


def test_version_arg():
    # GIVEN
    expected_version = smartchangelog.__version__
    with set_args(commitmsg_script_path, "--version") as result, pytest.raises(SystemExit) as e:
        # WHEN
        smartchangelog.scripts.commitmsg_script.main()
    stdout, stderr = result
    version = stdout.read().strip("\n")
    # THEN
    assert e.value.code == 0
    assert version == expected_version




import inspect

import pytest

import smartchangelog
from smartchangelog.scripts import changelog_script
from smartchangelog.tools import set_args
from smartchangelog.gitcmd import tag

"""Path of the file containing changelog_script.py file"""
changelog_script_path = inspect.getfile(changelog_script)


def test_range_arg():
    # GIVEN
    with set_args(changelog_script_path, '--range', revision_range()) as result, \
            pytest.raises(
                SystemExit) as e:
        # WHEN
        changelog_script.main()
    stdout, stderr = result
    printed_report = stdout.read()
    # THEN
    assert e.value.code == 0
    assert printed_report


def test_range_arg_with_groupby():
    # GIVEN
    with set_args(changelog_script_path, '--range', revision_range(), '--groupby', 'type', 'scope') as result, \
            pytest.raises(
                SystemExit) as e:
        # WHEN
        changelog_script.main()
    stdout, stderr = result
    printed_report = stdout.read()
    # THEN
    assert e.value.code == 0
    assert printed_report


def test_version_arg():
    # GIVEN
    expected_version = smartchangelog.__version__
    with set_args(changelog_script_path, "--version") as result, pytest.raises(SystemExit) as e:
        # WHEN
        changelog_script.main()
    stdout, stderr = result
    version = stdout.read().strip("\n")
    # THEN
    assert e.value.code == 0
    assert version == expected_version


# Tools
def revision_range() -> str:
    tags = tag()
    rev_range = "{start}..{end}".format(start=tags[0], end=tags[-1])
    return rev_range



from setuptools import setup, find_packages
from codecs import open
from os import path
from smartchangelog import __version__

here = path.abspath(path.dirname(__file__))

with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

name, description = long_description.split('\n')[0].split(': ')

setup(
    name=name,

    version=__version__,

    description=description,
    long_description=long_description,

    url='https://github.com/ngouzy/smartchangelog',

    author='Nicolas Gouzy',
    author_email='nicolas.gouzy@gmail.com',

    license='MIT',

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        'Development Status :: 4 - Beta',

        'Intended Audience :: Developers',
        'Topic :: Software Development :: Documentation',

        'License :: OSI Approved :: MIT License',

        # Supported Python versions
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3 :: Only',

        # Environment
        'Environment :: Console'
    ],

    keywords='changelog, git, hook, message formatter',

    packages=find_packages(exclude=['contrib', 'docs', 'tests']),

    install_requires=[],

    extras_require={
        'test': ['mypy', 'pytest', 'pytest-cov']
    },

    # To provide executable scripts, use entry points in preference to the
    # "scripts" keyword. Entry points provide cross-platform support and allow
    # pip to create the appropriate form of executable for the target platform.
    entry_points={
        'console_scripts': [
            'commit-msg=smartchangelog.scripts.commitmsg_script:main',
            'smartchangelog=smartchangelog.scripts.changelog_script:main',
        ],
    },
)



import inspect
import re
from enum import Enum

from typing import NamedTuple


class CommitSyntaxError(Exception):
    """
    Invalid commit syntax error
    """


class CommitType(Enum):
    feat = 'new feature for the user, not a new feature for build script'
    fix = 'bug fix for the user, not a fix to a build script'
    docs = 'changes to the documentation'
    style = 'formatting, missing semi colons, etc; no production code change'
    refactor = 'refactoring production code, eg.renaming a variable'
    test = 'adding missing tests, refactoring tests; no production code change'
    chore = 'updating gradle scripts, continuous integration scripts,  etc; no production code change'

    def __lt__(self, other):
        if isinstance(other, self.__class__):
            return self.index() < other.index()
        return NotImplemented

    def index(self):
        return [ct for ct in CommitType].index(self)

    def __str__(self):
        return self.name


class FirstLine(NamedTuple):
    type: CommitType
    scope: str
    subject: str


class CommitMsg:
    """
    Your commit message have to follow this format:
    <type>(<scope>): <subject>

    <body>

    <footer>
    Where :
    Message first line (type, scope and subject)
        The first line cannot be longer than {firstline_max_length} characters.
        The type and scope should always be lowercase as shown
        below.
        Allowed <type> values: {allowed_types}
        Example <scope> values:
            * ui
            * business
            * model
            * widget
            * config
            etc.
        The <scope> can be empty (e.g. if the change is a global or difficult
        to assign to a single component), in which case the parentheses are
        omitted.

    Message body (optional)
        If there is a body, it must have a blank line between the first line and
        the body.
        The body cannot be longer than {bodyline_max_length} characters.
        uses the imperative, present tense: "change" not "changed" nor
        "changes"
        includes motivation for the change and contrasts with previous behavior

    Message footer
        Referencing issues or user stories (Jira references)
        If there is a footer, it must have a body and it must have a blank line between the body and
        the footer.
        The footer cannot be longer than {footerline_max_length} characters.
    """
    FIRSTLINE_PATTERN = re.compile('^([a-z]+)(?:\(([^\n\t]+)\))?: (.+)$')
    FIRSTLINE_MAX_LENGTH = 70
    BODY_MAX_LENGTH = 80
    FOOTER_MAX_LENGTH = 80

    def __init__(self, msg_type: CommitType, scope: str, subject: str, body: str = None, footer: str = None) -> None:
        self.type = msg_type
        self.scope = scope
        self.subject = subject
        self.body = body
        self.footer = footer

    def __eq__(self, other) -> bool:
        if isinstance(other, self.__class__):
            return self.__dict__ == other.__dict__
        return NotImplemented

    @classmethod
    def parse(cls, msg: str) -> 'CommitMsg':
        msg_parts = msg.split("\n\n")
        firstline = cls.parse_firstline(msg_parts[0])
        if len(msg_parts) > 1:
            body = msg_parts[1]
            cls.parse_body(body)
        else:
            body = None
        if len(msg_parts) > 2:
            footer = msg_parts[2]
            cls.parse_footer(footer)
        else:
            footer = None
        return cls(firstline.type, firstline.scope, firstline.subject, body, footer)

    @classmethod
    def parse_firstline(cls, firstline: str) -> FirstLine:
        if len(firstline) > cls.FIRSTLINE_MAX_LENGTH:
            raise CommitSyntaxError("First line can not be greater than {length} characters".format(
                length=cls.FIRSTLINE_MAX_LENGTH))
        result = cls.FIRSTLINE_PATTERN.search(firstline)
        if "\n" in firstline.strip():
            raise CommitSyntaxError("Two blank lines have to separate the first line and body part")
        if result is None:
            raise CommitSyntaxError("{firstline} doesn't follow the first line commit message pattern: {pattern}"
                                    .format(firstline=firstline, pattern=cls.FIRSTLINE_PATTERN.pattern))
        commit_type_str, scope, subject = result.groups()
        try:
            commit_type = CommitType[commit_type_str]
        except KeyError:
            raise CommitSyntaxError("{commit_type} is not an available commit type".format(commit_type=commit_type_str))
        return FirstLine(type=commit_type, scope=scope, subject=subject)

    @classmethod
    def parse_body(cls, body: str) -> str:
        for line in body.split('\n'):
            if len(line) > cls.BODY_MAX_LENGTH:
                raise CommitSyntaxError("Body line can not be greater than {length} characters".format(
                    length=cls.BODY_MAX_LENGTH))
        return body

    @classmethod
    def parse_footer(cls, footer: str) -> str:
        for line in footer.split('\n'):
            if len(line) > cls.FOOTER_MAX_LENGTH:
                raise CommitSyntaxError("Footer line can not be greater than {length} characters".format(
                    length=cls.FOOTER_MAX_LENGTH))
        return footer

    @classmethod
    def format_allowed_types(cls) -> str:
        return "\n" + "\n".join("\t* {name}: {doc}".format(name=ct.name, doc=ct.value) for ct in CommitType)

    @classmethod
    def help(cls) -> str:
        return inspect.getdoc(cls).format(allowed_types=cls.format_allowed_types(),
                                          firstline_max_length=cls.FIRSTLINE_MAX_LENGTH,
                                          bodyline_max_length=cls.BODY_MAX_LENGTH,
                                          footerline_max_length=cls.FOOTER_MAX_LENGTH)

